from collections.abc import Sequence as SequenceType
from typing import Any, Union, Callable, Iterable, cast, Collection, List, Tuple

import pytest


TestArgument = Union[str, List[str], Tuple[str, ...]]  # can also be a comma-separated string
TestIdFormatter = Union[Callable[[Any], str], Collection[str]]
TestFunction = Callable[..., None]


def _parametrized_test_formatter(test_case_value: Any) -> str:
    """ produce a prettier version of a parameter id """
    if isinstance(test_case_value, str):
        pretty = test_case_value
    elif isinstance(test_case_value, SequenceType):
        pretty = "-".join(str(test_case) for test_case in test_case_value)
    else:
        pretty = str(test_case_value)
    return pretty.replace(".", "-")


def parametrize(
    arguments: TestArgument,
    value: Iterable[Any],
    ids: TestIdFormatter = _parametrized_test_formatter,
    **kwargs: Any
) -> TestFunction:
    """ augment the pytest decorator with some id magic (bonus: no more 'pytest.mark.parametrize' typos!) """
    return cast(TestFunction, pytest.mark.parametrize(arguments, value, ids=ids, **kwargs).__call__)



""" Various testing tools involving loggers. """

from contextlib import contextmanager
import logging
from io import StringIO
from types import TracebackType
from typing import Generator, Iterable, Optional


StringCollection = Iterable[str]


class SetLoggingLevel:
    """
    Class used to change the logging level for a block of code.

    For example, we can change the logging level for 'elasticsearch' to ERROR while calling DoSomething:

            with SetLoggingLevel('elasticsearch', logging.ERROR):
                DoSomething()

    """

    def __init__(self, logger_name: str, logging_level: int = logging.CRITICAL) -> None:
        """
        Initializer.

        :param logger_name: The logger name.
        :param logging_level: The temporary logging level.
        """
        self._logger = logging.getLogger(logger_name)  # Create logger if it does not exists yet.
        assert self._logger
        self._logging_level = logging_level
        self._old_logging_level: Optional[int] = None

    def __enter__(self) -> None:
        """
        Change the logging level.
        """
        self._old_logging_level = self._logger.getEffectiveLevel()
        self._logger.setLevel(self._logging_level)

    # noinspection PyUnusedLocal
    def __exit__(
        self,
        _exc_type: Optional[type],
        _exc_val: Optional[Exception],
        _exc_tb: Optional[TracebackType],
    ) -> None:
        """
        Restore the logging level to its original value.
        """
        self._logger.setLevel(self._old_logging_level)
        assert self._logger.getEffectiveLevel() == self._old_logging_level


@contextmanager
def intercept_logging(logger: logging.Logger, level: int = None) -> Generator[StringIO, None, None]:
    """
    Context manager. Add a temporary stream handler to a logger.

    :param logger: a python logger
    :param level: a python log level
    :return: yields the stream (use .get_value() to extract the stuff!)
    """
    __tracebackhide__ = True
    stream = StringIO()
    handler = logging.StreamHandler(stream)
    logger.addHandler(handler)

    with SetLoggingLevel(logger.name, level or logger.level):
        yield stream
        handler.flush()

    logger.removeHandler(handler)


@contextmanager
def assert_logging(
    logger: logging.Logger,
    present: StringCollection = None,
    absent: StringCollection = None,
    level: int = None,
) -> Generator[None, None, None]:
    """
    Context manager that attaches to a logger and ensures the presence of specific strings when the context is left.
    An assert will be raised as soon as a string from `intercepts` is missing from the logs.

    :param logger: a python logger
    :param present: An iterable of strings to find in the logs (case sensitive)
    :param absent: An iterable of strings that must NOT be found in the logs (case sensitive)
    :param level: specify a logging level for this handler only.
    """
    __tracebackhide__ = True
    if isinstance(present, str):
        present = (present,)

    if isinstance(absent, str):
        absent = (absent,)

    with intercept_logging(logger, level) as stream:
        try:
            yield
        finally:
            logs = stream.getvalue()
            if present:
                assert logs.strip(), "log output was silent"
                for line in present:
                    assert line in logs, f'String "{line}" cannot be found in the logs:\n{logs}'
            if absent:
                for line in absent:
                    assert line not in logs, f'String "{line}" was found in the logs:\n{logs}'



from abc import abstractmethod
from contextlib import contextmanager
import logging
from typing import Generator, TypeVar


log = logging.getLogger(__name__)

# Type that represent the type of self for classes that derives from TmpResource.
T = TypeVar("T", bound="TemporaryResource")


class TemporaryResource:
    """ Base class used for creating temporary unit test objects """

    _disposed: bool = False

    @abstractmethod
    def create_resource(self) -> None:
        """ Creates the object. """
        assert not self._disposed

    @abstractmethod
    def delete_resource(self) -> None:
        """ Delete the temporary object(s) """
        assert not self._disposed
        self._disposed = True

    @contextmanager
    def auto_delete(self: T) -> Generator[T, None, None]:
        """ Provides a contextmanager that deletes the resource on exit. """
        assert not self._disposed
        yield self
        self.delete_resource()



from datetime import datetime
import itertools
import os
import platform
import re
import threading
from typing import Dict, ClassVar, Iterator

import pytest
from _pytest.fixtures import SubRequest
from typing_extensions import Final


class TestId:
    """
    Generates a "unique-enough" test id made up from human-readable components.

    - Threaded test executors will increment sequence number.
    - Multiprocess test executors would use different pids.
    - Jenkins executors provide different executor labels.
    """

    DELIMITER: Final[str] = "."
    HOST: Final[str] = platform.node()
    PID: Final[int] = os.getpid()
    TIMESTAMP: Final[datetime] = datetime.now()
    # this exists on jenkins. using it ensures that running parallel tests on shared resources (like docker sockets)
    # cannot create the same id.
    EXECUTOR: Final[str] = os.environ.get("EXECUTOR_NUMBER", "default")

    # each friendly name keeps a count.
    _sequence_count: ClassVar[Dict[str, Iterator[int]]] = {}
    _threading_lock: ClassVar[threading.Lock] = threading.Lock()

    def __init__(self, friendly_name: str) -> None:
        self.friendly_name: str = friendly_name
        with self._threading_lock:
            self.sequence = next(self._sequence_count.setdefault(friendly_name, itertools.count()))
        parts = tuple(
            map(
                self._isolate_and_sanitize_id_part,
                (
                    self.friendly_name,
                    self.TIMESTAMP.strftime("%m%d%H%M%S"),
                    str(self.PID),
                    self.HOST,
                    self.EXECUTOR,
                    str(self.sequence),
                ),
            )
        )
        assert self.DELIMITER not in "".join(parts)
        self.id = self.DELIMITER.join(parts)

    @classmethod
    def _isolate_and_sanitize_id_part(cls, string: str) -> str:
        """ Return a new string that doesn't contain invalid characters or the name tokenization delimiter """
        replacement = "-"
        assert (
            replacement != cls.DELIMITER
        ), "Use a replacement character that's different from the delimiter."
        # remove delimiter from pattern so that it disappears from the string
        replace_pattern = r"[^a-zA-Z0-9_.-]".replace(cls.DELIMITER, "")
        return re.sub(replace_pattern, replacement, string)

    def __str__(self) -> str:
        return self.id


@pytest.fixture
def unique_test_id(request: SubRequest) -> TestId:
    """Provide a TestId object scoped to your test method

    Parametrized methods will trigger one TestId per parameter. The TestId's friendly name will include the "ids"
    that pytest generates for each parametrized test.

    e.g.:
        @parametrize('value', (3, True))
        def test_params(unique_test_id: TestId, value: int) -> None:
            assert unique_test_id.friendly_name == f"test_params[{value}]"  # test_params[3], test_params[True], etc
            assert unique_test_id.sequence == 0

    e.g.:
        @parametrize('value', (3, True), ids=('three', 'bool'))
        def test_params(unique_test_id: TestId, value: Any) -> None:
            ...  # test_params[three], test_params[bool], etc
    """
    return TestId(request.node.name)



import pytest
from _pytest.config import Config
from _pytest.mark import MarkDecorator

UnitTest = pytest.mark.unit_test  # :sofast:
Integration = pytest.mark.integration  # interacts with the product
Interactive = (
    pytest.mark.interactive
)  # it won't work OOTB / manual steps / eye-based validations...

# Registering the markers removes pytest warnings.
AUTO_REGISTER_TEST_MARKERS = [UnitTest, Integration, Interactive]


def register_markers(config: Config, *additional_markers: MarkDecorator) -> None:
    """
    Auto registers markers into pytest to silence some warnings.

    Must be bootstrapped from conftest.py:

        def pytest_configure(config: Config) -> None:
            register_markers(config)
    """
    for marker in AUTO_REGISTER_TEST_MARKERS + list(additional_markers):
        config.addinivalue_line("markers", f"env({marker.name})")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['coveo_testing', 'coveo_testing.temporary_resource']

package_data = \
{'': ['*']}

install_requires = \
['attrs', 'pytest']

setup_kwargs = {
    'name': 'coveo-testing',
    'version': '1.0.1',
    'description': 'Lightweight testing helpers',
    'long_description': '# coveo-testing\n\nA set of test/pytest helpers to facilitate common routines.\n\n\nContent in a nutshell:\n\n- Reusable pytest markers (UnitTest, IntegrationTest)\n- Unique ID generation for tests\n- Multiline logging assertions with includes, excludes, levels and comprehensive assertion output\n- Refactorable `unittest.mock.patch(\'this.module\')` module references\n\n- Human-readable (but still customizable) display for parametrized tests\n\n\nThis project is used as the test base for all other projects in this repository.\n\nTherefore, it cannot depend on any of them.\n\nMore complex use cases may be implemented in the `coveo-testing-extras` project. That\'s also where you can depend on projects that depend on `coveo-testing`. \n\n\n# pytest markers and auto-registration\n\nThis enables code completion on markers.\n\nThree markers are already provided: `[UnitTest, Integration, Interactive]`\n\nHere\'s how you can create additional markers:\n\n```python\n# /test_some_module/markers.py\nimport pytest\n\nDockerTest = pytest.mark.docker_test\nCloudTest = pytest.mark.cloud_test\n\nALL_MARKERS = [DockerTest, CloudTest]\n```\n\nYou can then import these markers and decorate your test functions accordingly:\n\n```python\n# /test_some_module/test_something.py\nfrom coveo_testing.markers import UnitTest, Integration, Interactive\nfrom test_some_module.markers import CloudTest, DockerTest\n\n@UnitTest\ndef test_unit() -> None:\n    ...  # designed to be fast and lightweight, most likely parametrized\n\n\n@Integration\ndef test_integration() -> None:\n    ...  # combines multiple features to achieve a test\n\n\n@CloudTest\ndef test_in_the_cloud() -> None:\n    ...  # this could be a post-deployment test, for instance.\n\n\n@DockerTest\n@Integration\ndef test_through_docker() -> None:\n    ... # will run whenever docker tests or integration tests are requested\n\n\n@Interactive\ndef test_interactive() -> None:\n    ...  # these tests rely on eye-validations, special developer setups, etc  \n\n```\n\nPytest will issue a warning when markers are not registered.\n\nTo register coveo-testing\'s markers along with your custom markers, use the provided `register_markers` method:\n\n```python\n# /test_some_module/conftest.py\nfrom _pytest.config import Config\nfrom coveo_testing.markers import register_markers\nfrom test_some_module.markers import ALL_MARKERS\n\ndef pytest_configure(config: Config) -> None:\n    """This pytest hook is ran once, before collecting tests."""\n    register_markers(config, *ALL_MARKERS)\n```\n\n\n# Human-readable unique ID generation\n\nThe generated ID has this format:\n\n`friendly-name.timestamp.pid.host.executor.sequence`\n\n- friendly-name:\n  - provided by you, for your own benefit\n    \n- timestamp: \n  - format "%m%d%H%M%S" (month, day, hour, minutes, seconds)\n  - computed once, when TestId is imported\n    \n- pid:\n  - the pid of the python process\n    \n- host:\n  - the network name of the machine\n\n- executor:\n  - the content of the `EXECUTOR_NUMBER` environment variable\n  - returns \'default\' when not defined  \n  - historically, this variable comes from jenkins\n  - conceptually, it can be used to help distribute (and identify) tests and executors\n\n- sequence:\n  - Thread-safe\n  - Each `friendly-name` has an isolated `sequence` that starts at 0\n  - Incremented on each new instance\n  - Enables support for parallel parametrized tests\n\n```python\nfrom coveo_testing.temporary_resource.unique_id import TestId, unique_test_id\n\n\n# the friendly name is the only thing you need to specify\ntest_id = TestId(\'friendly-name\')\nstr(test_id)\n\'friendly-name.0202152243.18836.WORKSTATION.default.0\'\n\n\n# you can pass the instance around to share the ID\nstr(test_id)\n\'friendly-name.0202152243.18836.WORKSTATION.default.0\'\n\n\n# create new instances to increment the sequence number\ntest_id = TestId(\'friendly-name\')\nstr(test_id)\n\'friendly-name.0202152243.18836.WORKSTATION.default.1\'\n\n\n# use it in parallel parameterized tests\nimport pytest\n\n@pytest.mark.parametrize(\'param\', (True, False))\ndef test_param(param: bool, unique_test_id: TestId) -> None:\n    # in this case, the friendly name is the function name and\n    # the sequence will increase on each parameter\n    # test_param.0202152243.18836.WORKSTATION.default.0\n    # test_param.0202152243.18836.WORKSTATION.default.1\n    ...\n```\n\n\n# multiline logging assertions\n\nMaybe pytest\'s `caplog` is enough for your needs, or maybe you need more options.\nThis tool uses `in` and `not in` to match strings in a case-sensitive way.\n\n```python\nimport logging\nfrom coveo_testing.logging import assert_logging\n\nwith assert_logging(\n        logging.getLogger(\'logger-name\'),\n        present=[\'evidence1\', \'evidence2\'], \n        absent=[...], \n        level=logging.WARN):\n    ...\n```\n\n\n# Human-readable (but still customizable) display for parametrized tests\n\nIf you\'re like me, you typed `@pytest.mark.parametrize` wrong a couple of times!\n\nEnable IDE completion by using this one instead:\n\n```python\nfrom coveo_testing.parametrize import parametrize\n\n@parametrize(\'var\', (True, False))\ndef test_var(var: bool) -> None:\n    ...\n```\n\nIt has one difference vs the pytest one, and it\'s the way it formats the "parameter name" for each iteration of the test.\n\nPytest will skip a lot of types and will simply name your test "var0", "var1" and so on.\nUsing this `@parametrize` instead, the variable\'s content will be inspected:\n\n```python\nfrom typing import Any\nfrom coveo_testing.parametrize import parametrize\nimport pytest\n\n\nclass StrMe:\n  def __init__(self, var: Any) -> None:\n    self.var = var\n      \n  def __str__(self) -> str:\n    return f"Value: {self.var}"\n\n\n@parametrize(\'var\', [[\'list\', \'display\'], [StrMe(\'hello\')]])\ndef test_param(var: bool) -> None:\n    ...\n\n@pytest.mark.parametrize(\'var\', [[\'list\', \'display\'], [StrMe(\'hello\')]])\ndef test_param_from_pytest(var: bool) -> None:\n    ...\n```\n\nIf you run `pytest --collect-only` you will obtain the following:\n```\n    <Function test_param[list-display]>\n    <Function test_param[Value: hello]>\n    <Function test_param_from_pytest[var0]>\n    <Function test_param_from_pytest[var1]>\n```\n',
    'author': 'Jonathan Piché',
    'author_email': 'tools@coveo.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/coveooss/coveo-python-oss',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)



from typing import Any


def resolve_mock_target(target: Any) -> str:
    """
    `mock.patch` uses a str-representation of an object to find it, but this doesn't play well with
    refactors and renames. This method extracts the str-representation of an object.

    This method will not handle _all_ kinds of objects, in which case an AttributeError will most likely be raised.
    """
    return f"{target.__module__}.{target.__name__}"
